@@ -12,6 +12,7 @@ __all__ = """
     Size
     set_autodevice
     unset_autodevice
+    is_autodevice
 """.split()
 
 try:
@@ -28,22 +29,30 @@ from pyctlib import raw_function, return_type_wrapper, touch
 from functools import wraps
 from types import GeneratorType, MethodWrapperType
 from collections import OrderedDict
+from .torch_namespace import *
+from .device import AutoDevice as Device
 
 """
 from torchplus import Tensor
 import torchplus as tp
 """
 
-Device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
+# Device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
 
 _auto_device = True
 
 def set_autodevice(flag=True):
+    global _auto_device
     _auto_device = flag
 
 def unset_autodevice():
+    global _auto_device
     _auto_device = False
 
+def is_autodevice():
+    global _auto_device
+    return _auto_device
+
 # def totensor(x) -> 'Tensor':
 #     if isinstance(x, Tensor):
 #         return x
@@ -164,10 +173,6 @@ class Size(tuple):
     @property
     def bc(self): return [x for x in [self.batch_dimension, self.channel_dimension] if x is not None]
 
-    def special_from(self, other):
-        self.batch_dimension = getattr(other, '_batch_dimension', None)
-        self.channel_dimension = getattr(other, '_channel_dimension', None)
-
     def special_from_(self, other):
         self.batch_dimension = getattr(other, '_batch_dimension', None)
         self.channel_dimension = getattr(other, '_channel_dimension', None)
@@ -176,9 +181,9 @@ class Size(tuple):
     @property
     def space(self):
         s = self.special
-        if len(s) == 0: return self.remove_special()
-        elif len(s) == 1: return (self[:s[0]] + self[s[0]+1:]).remove_special()
-        return (self[:s[0]] + self[s[0]+1:s[1]] + self[s[1]+1:]).remove_special()
+        if len(s) == 0: return self
+        elif len(s) == 1: return (self[:s[0]] + self[s[0]+1:])
+        return (self[:s[0]] + self[s[0]+1:s[1]] + self[s[1]+1:])
 
     @property
     def nele(self):
@@ -245,7 +250,6 @@ class Size(tuple):
     @params
     @staticmethod
     def __op__(a: [builtins.int, tuple, 'Size'], b: [builtins.int, tuple, 'Size'], *, op):
-        print(a, b)
         if not isinstance(a, Size): a = Size(a, force=True)
         if not isinstance(b, Size): b = Size(b, force=True)
         def getvalue(x, y):
@@ -454,9 +458,11 @@ class Size(tuple):
         if isinstance(k, builtins.int): return super().__getitem__(k)
         if isinstance(k, slice):
             s = Size(super().__getitem__(k))
+            if k.start is None: ks = 0
+            if k.start is not None and k.start < 0: ks = k.start + self.ndim
             try:
-                s.batch_dimension = self.batch_dimension - k.start
-                s.channel_dimension = self.channel_dimension - k.start
+                s.batch_dimension = self.batch_dimension - ks
+                s.channel_dimension = self.channel_dimension - ks
             except: pass
             return s
         return Size(self.python_repr[k])
@@ -485,6 +491,8 @@ class Tensor(torch.Tensor):
         self = super()._make_subclass(cls, data, requires_grad)
         if device is None and auto_device: return self.to(Device)
         if device is not None: return self.to(device)
+        if isinstance(data, cls):
+            self.special_from_(data)
         return self
 
     def __new__(cls, *args, **kwargs):
@@ -515,6 +523,13 @@ class Tensor(torch.Tensor):
     @classmethod
     def __torch_function__(cls, func, types, args=(), kwargs=None):
         self = args[0]
+        assert isinstance(self, torch.Tensor)
+        if type(self) != Tensor:
+            isTensor = [type(x) == Tensor for x in args]
+            isofTensor = [isinstance(x, Tensor) for x in args]
+            if builtins.any(isTensor): self = args[isTensor.index(True)]
+            elif builtins.any(isofTensor): self = args[isofTensor.index(True)]
+            else: return super().__torch_function__(func, types, args, kwargs)
         if func == torch.Tensor.dim:
             return super().__torch_function__(func, types, args, kwargs)
         ibatch = self.batch_dimension
@@ -524,35 +539,65 @@ class Tensor(torch.Tensor):
             torch.reshape, torch.Tensor.reshape, torch.Tensor.view,
             torch.zeros, torch.ones, torch.rand, torch.randn
         ]:
-            dims = Size(args[1:])
-            args = args[:1] + dims
-        if func in [torch.unsqueeze, torch.Tensor.unsqueeze, torch.Tensor.unsqueeze_]:
             dims = args[1:]
+            if len(dims) == 1: dims = dims[0]
+            args = args[:1] + Size(dims)
         elif func in [torch.Tensor.view_as, torch.Tensor.reshape_as]:
             dims = args[1].shape
         elif func == torch.randint:
             dims = Size(args[3])
             args = args[:3] + (dims,)
-        types = tuple(cls if t == torch.nn.Parameter else t for t in types)
+        types = tuple(cls if t in [torch.nn.Parameter, tp.nn.Parameter] else t for t in types)
         ret = super().__torch_function__(func, types, args, kwargs)
         if isinstance(ret, type(NotImplemented)):
             raise NotImplementedError(f"{func} for {args} is not implemented. ")
-        if not isinstance(self, torch.Tensor): return ret
         if not isinstance(self, Tensor): self = Tensor(self)
-        to_squeeze = False
-        if not isinstance(ret, tuple):
-            ret = (ret,)
-            to_squeeze = True
-        for r in ret:
-            if not isinstance(r, torch.Tensor): continue
-            if not isinstance(r, Tensor): r = Tensor(r)
+        def _convert(ret, cls):
+            if isinstance(ret, torch.Tensor):
+                ret = ret.as_subclass(cls)
+            if isinstance(ret, (tuple, list)):
+                # Also handles things like namedtuples
+                ret = type(ret)(_convert(r, cls) for r in ret)
+            return ret
+        def _collect(r, c):
+            if isinstance(r, (tuple, list)):
+                for x in r: _collect(x, c)
+            if isinstance(r, torch.Tensor): c.append(r)
+        collection = []
+        ret = _convert(ret, Tensor)
+        _collect(ret, collection)
+        for r in collection:
             if func in [
                 torch.reshape, torch.Tensor.reshape, torch.Tensor.reshape_as,
                 torch.Tensor.view, torch.Tensor.view_as,
                 torch.zeros, torch.ones, torch.rand, torch.randn, torch.randint
             ]:
-                r.special_from(dims)
+                r.special_from_(dims)
             if (ibatch is not None or ichannel is not None) and not r.has_special:
+                if func in [torch.transpose, torch.Tensor.transpose, torch.Tensor.transpose_]:
+                    dim1, dim2 = args[1:]
+                    if dim1 < 0: dim1 += ndim
+                    if dim2 < 0: dim2 += ndim
+                    r._batch_dimension = ibatch
+                    r._channel_dimension = ichannel
+                    if ibatch == dim1: r._batch_dimension = dim2
+                    elif ibatch == dim2: r._batch_dimension = dim1
+                    if ichannel == dim1: r._channel_dimension = dim2
+                    elif ichannel == dim2: r._channel_dimension = dim1
+                    continue
+                if func == torch.Tensor.permute:
+                    dims = args[1:]
+                    if ibatch is not None: r._batch_dimension = dims.index(ibatch)
+                    if ichannel is not None: r._channel_dimension = dims.index(ichannel)
+                    continue
+                if func in [torch.mm, torch.bmm, torch.smm]:
+                    self, other = args[:2]
+                    r.special_from_(self.shape[:-1] + other.shape[-1:])
+                    continue
+                if func in [torch.addmm, torch.addbmm, torch.saddmm]:
+                    self, other = args[1:]
+                    r.special_from_(self.shape[:-1] + other.shape[-1:])
+                    continue
                 if ndim == len(r.ishape):
                     r._batch_dimension = ibatch
                     r._channel_dimension = ichannel
@@ -570,27 +615,63 @@ class Tensor(torch.Tensor):
                         if len(dims) == 1 and iterable(dims[0]): dims = dims[0]
                         if ibatch in dims: ibatch = None
                         if ichannel in dims: ichannel = None
-                        if ibatch is not None: r._batch_dimension = ibatch - len([d for d in dims if d < ibatch])
-                        if ichannel is not None: r._channel_dimension = ichannel - len([d for d in dims if d < ichannel])
+                        if ibatch is not None: r._batch_dimension = ibatch - len([d for d in dims if 0 <= d < ibatch or d + ndim < ibatch])
+                        if ichannel is not None: r._channel_dimension = ichannel - len([d for d in dims if 0 <= d < ichannel or d + ndim < ichannel])
                 elif func in [torch.squeeze, torch.Tensor.squeeze, torch.Tensor.squeeze_]:
                     dims = args[1:]
                     if len(dims) == 1 and iterable(dims[0]): dims = dims[0]
                     if len(dims) == 0: dims = tuple(i for i, x in enumerate(self.ishape) if x == 1)
                     if ibatch in dims: ibatch = None
                     if ichannel in dims: ichannel = None
-                    if ibatch is not None: r._batch_dimension = ibatch - len([d for d in dims if d < ibatch])
-                    if ichannel is not None: r._channel_dimension = ichannel - len([d for d in dims if d < ichannel])
+                    if ibatch is not None: r._batch_dimension = ibatch - len([d for d in dims if 0 <= d < ibatch or d + ndim < ibatch])
+                    if ichannel is not None: r._channel_dimension = ichannel - len([d for d in dims if 0 <= d < ichannel or d + ndim < ichannel])
                 elif func in [torch.unsqueeze, torch.Tensor.unsqueeze, torch.Tensor.unsqueeze_]:
                     dims = args[1:]
                     if len(dims) == 1 and iterable(dims[0]): dims = dims[0]
                     for d in dims:
-                        if ibatch is not None and d <= ibatch: ibatch += 1
-                        if ichannel is not None and d <= ichannel: ichannel += 1
+                        if ibatch is not None and (0 <= d <= ibatch or d + ndim <= ibatch): ibatch += 1
+                        if ichannel is not None and (0 <= d <= ichannel or d + ndim <= ichannel): ichannel += 1
                     if ibatch is not None: r._batch_dimension = ibatch
                     if ichannel is not None: r._channel_dimension = ichannel
+                elif func in [torch.flatten, torch.Tensor.flatten]:
+                    dims = args[1:]
+                    if len(dims) == 1:
+                        dim = dims[0]
+                        if dim < 0: dim += ndim
+                        if ibatch is not None and ibatch >= dim: r._batch_dimension = None
+                        if ichannel is not None and ichannel >= dim: r._channel_dimension = None
+                    else:
+                        dim1, dim2 = dims
+                        if dim1 < 0: dim1 += ndim
+                        if dim2 < 0: dim2 += ndim
+                        if ibatch is not None and dim1 <= ibatch <= dim2: r._batch_dimension = None
+                        if ichannel is not None and dim1 <= ichannel <= dim2: r._channel_dimension = None
+                elif func == torch.Tensor.__getitem__:
+                    dims = args[1]
+                    if not isinstance(dims, tuple): dims = (dims,)
+                    if ... in dims:
+                        lp = dims[:dims.index(...)]
+                        rp = dims[dims.index(...)+1:]
+                    else:
+                        lp = dims
+                        rp = tuple()
+                    offset = ndim - len(rp)
+                    if ibatch is not None:
+                        if ibatch < len(lp) and isinstance(lp[ibatch], builtins.int): ibatch = None
+                        elif ibatch >= offset and isinstance(rp[ibatch - offset], builtins.int): ibatch = None
+                    if ichannel is not None:
+                        if ichannel < len(lp) and isinstance(lp[ichannel], builtins.int): ichannel = None
+                        elif ichannel >= offset and isinstance(rp[ichannel - offset], builtins.int): ichannel = None
+                    if ibatch is not None:
+                        r._batch_dimension = ibatch - \
+                            len([d for d in builtins.range(len(lp)) if (0 <= d < ibatch or d + ndim < ibatch) and isinstance(lp[d], builtins.int)]) - \
+                            len([d for d in builtins.range(offset, ndim) if (0 <= d < ibatch or d + ndim < ibatch) and isinstance(rp[d-offset], builtins.int)])
+                    if ichannel is not None:
+                        r._channel_dimension = ichannel - \
+                            len([d for d in builtins.range(len(lp)) if (0 <= d < ichannel or d + ndim < ichannel) and isinstance(lp[d], builtins.int)]) - \
+                            len([d for d in builtins.range(offset, ndim) if (0 <= d < ichannel or d + ndim < ichannel) and isinstance(rp[d-offset], builtins.int)])
                 else: raise RuntimeError(f"{func} needs to be override. ")
-        if to_squeeze: return ret[0]
-        else: return ret
+        return ret
 
     def requires_grad_(self, mode=True):
         self.requires_grad = mode
@@ -600,10 +681,6 @@ class Tensor(torch.Tensor):
         self.has_names = True
         return super().refine_names(*args, **kwargs)
 
-    def special_from(self, other):
-        self.batch_dimension = getattr(other, '_batch_dimension', None)
-        self.channel_dimension = getattr(other, '_channel_dimension', None)
-
     def special_from_(self, other):
         self.batch_dimension = getattr(other, '_batch_dimension', None)
         self.channel_dimension = getattr(other, '_channel_dimension', None)
@@ -777,20 +854,57 @@ class Tensor(torch.Tensor):
         if dim is None or isinstance(dim, list) and dim == []: dim = self.batch_dimension
         if dim is None or isinstance(dim, set) and dim == {}: dim = self.channel_dimension
         if dim is None: raise TypeError("Argument 'dim' needed for sampling Tensors with no special dimensions. ")
+        if number < 1: raise TypeError("Argument 'number' for sampling Tensors can not be smaller than 1. ")
         sample_indices = [slice(None)] * self.dim()
         if self.shape[dim] < number: raise TypeError(f"Too many elements needed to be sampled from dimension {dim}")
         if random:
             import random
             samples = random.sample(builtins.range(self.shape[dim]), k = number)
         else: samples = list(builtins.range(number))
-        sample_indices[dim] = samples
+        sample_indices[dim] = samples if len(samples) > 1 else samples[0]
         output_tensor = Tensor(self[tuple(sample_indices)])
         output_tensor.indices = samples
         return output_tensor
 
+    @params
+    def pick(self, dim: builtins.int, index: builtins.int):
+        """
+        pick(self, dim, index) -> Tensor
+
+        pick one of the item on dimension `dim` for big tensors. 
+        data.pick(2, 4) is equivalent to data[:, :, 4]
+        """
+        if dim < 0: dim += self.ndim
+        return self[(slice(None),) * dim + (index,)]
+
+    @params
+    def mvdim(self, dim1: builtins.int, dim2: builtins.int):
+        """
+        mvdim(self, dim1, dim2) -> Tensor
+
+        move dim1 to dim2(specified in the targeting size)
+        data of size (2, 3, 4, 5) can be transform to (2, 4, 5, 3) by data.mvdim(1, -1) or data.mvdim(1, 3)
+        """
+        if dim1 < 0: dim1 += self.ndim
+        if dim2 < 0: dim2 += self.ndim
+        if dim1 == dim2: return self
+        elif dim1 < dim2: return self.unsqueeze(dim2+1).transpose(dim1, dim2+1).squeeze(dim1)
+        else: return self.unsqueeze(dim2).transpose(dim1+1, dim2).squeeze(dim1+1)
+
+    @params
+    def mvdim_(self, dim1: builtins.int, dim2: builtins.int):
+        """
+        In-place operation for mvdim
+        """
+        if dim1 < 0: dim1 += self.ndim
+        if dim2 < 0: dim2 += self.ndim
+        if dim1 == dim2: return self
+        elif dim1 < dim2: return self.unsqueeze_(dim2+1).transpose_(dim1, dim2+1).squeeze_(dim1)
+        else: return self.unsqueeze_(dim2).transpose_(dim1+1, dim2).squeeze_(dim1+1)
+
     @property
     def T(self: 'Tensor') -> 'Tensor':
-        if not self.has_special: return super().T
+        if not self.has_special: return Tensor(super().T)
         s = self.special
         if len(s) == 1: permute_dim = tuple(builtins.range(s[0]))[::-1] + (s[0],) + tuple(builtins.range(s[0]+1, self.ndim))[::-1]
         elif len(s) == 2: permute_dim = tuple(builtins.range(s[0]))[::-1] + (s[0],) + tuple(builtins.range(s[0]+1, s[1]))[::-1] + (s[1],) + tuple(builtins.range(s[1]+1, self.ndim))[::-1]
@@ -1002,16 +1116,19 @@ for key in dir(torch):
         continue
     if key in __all__ or key in globals():
         continue
-    if isinstance(eval(f"torch.{key}"), torch.dtype):
+    if key in torch_dtype_list:
         # dtypes
         exec(f"{key} = torch.{key}")
         __all__.append(key)
-    if key not in {"optim", "random"}:
+    if key in torch_func_list:
         # functions
         exec(f"{key} = torch.{key}")
         __all__.append(key)
+    if key in torch_type_list:
+        exec(f"{key} = torch.{key}")
+        __all__.append(key)
 
-for key in '''
-no_grad optim random
-'''.split():
-    exec(f"{key} = torch.{key}")
+# for key in '''
+# no_grad
+# '''.split():
+#     exec(f"{key} = torch.{key}")
